@@ -858,6 +858,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 				         int skip_reply,
 				         ARGV *macros,...)
 {
+    const char *myname = "milter8_event";
     va_list ap;
     ssize_t data_len;
     int     err;
@@ -871,6 +872,17 @@ static const char *milter8_event(MILTER8 *milter, int event,
 
 #define DONT_SKIP_REPLY	0
 
+    /*
+     * Sanity check.
+     */
+    if (milter->fp == 0 || milter->def_reply != 0) {
+	msg_warn("%s: attempt to send event %s to milter %s after error",
+		 myname,
+		 (smfic_name = str_name_code(smfic_table, event)) != 0 ?
+		 smfic_name : "(unknown MTA event)", milter->m.name);
+	return (milter->def_reply);
+    }
+
     /*
      * Skip this event if it doesn't exist in the protocol that I announced.
      */
@@ -1185,7 +1197,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 						       (ssize_t) index,
 						       STR(milter->buf));
 		    if (edit_resp)
-			return (milter8_def_reply(milter, edit_resp));
+			return (edit_resp);
 		    continue;
 #endif
 
@@ -1203,7 +1215,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 						   STR(milter->buf),
 						   STR(milter->body));
 		    if (edit_resp)
-			return (milter8_def_reply(milter, edit_resp));
+			return (edit_resp);
 		    continue;
 
 		    /*
@@ -1232,7 +1244,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 						   STR(milter->buf),
 						   STR(milter->body));
 		    if (edit_resp)
-			return (milter8_def_reply(milter, edit_resp));
+			return (edit_resp);
 		    continue;
 #endif
 
@@ -1248,7 +1260,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    edit_resp = parent->add_rcpt(parent->chg_context,
 						 STR(milter->buf));
 		    if (edit_resp)
-			return (milter8_def_reply(milter, edit_resp));
+			return (edit_resp);
 		    continue;
 
 		    /*
@@ -1263,7 +1275,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    edit_resp = parent->del_rcpt(parent->chg_context,
 						 STR(milter->buf));
 		    if (edit_resp)
-			return (milter8_def_reply(milter, edit_resp));
+			return (edit_resp);
 		    continue;
 
 		    /*
@@ -1280,7 +1292,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    edit_resp = parent->repl_body(parent->chg_context,
 						  milter->body);
 		    if (edit_resp)
-			return (milter8_def_reply(milter, edit_resp));
+			return (edit_resp);
 		    continue;
 #endif
 		}
@@ -1927,6 +1939,16 @@ static void milter8_header(void *ptr, int unused_header_class,
     char   *cp;
     int     skip_reply;
 
+    /*
+     * XXX Workaround: mime_state_update() may invoke multiple call-backs
+     * before returning to the caller.
+     */
+#define MILTER8_MESSAGE_DONE(milter, msg_ctx) \
+	((milter)->state != MILTER8_STAT_MESSAGE || (msg_ctx)->resp != 0)
+
+    if (MILTER8_MESSAGE_DONE(milter, msg_ctx))
+	return;
+
     /*
      * XXX Sendmail compatibility. Don't expose our first (received) header
      * to mail filter applications. See also cleanup_milter.c for code to
@@ -1990,6 +2012,8 @@ static void milter8_eoh(void *ptr)
     MILTER_MSG_CONTEXT *msg_ctx = (MILTER_MSG_CONTEXT *) ptr;
     MILTER8 *milter = msg_ctx->milter;
 
+    if (MILTER8_MESSAGE_DONE(milter, msg_ctx))
+	return;
     if (msg_verbose)
 	msg_info("%s: eoh milter %s", myname, milter->m.name);
     msg_ctx->resp =
@@ -2012,6 +2036,9 @@ static void milter8_body(void *ptr, int rec_type,
     ssize_t space;
     ssize_t count;
 
+    if (MILTER8_MESSAGE_DONE(milter, msg_ctx))
+	return;
+
     /*
      * XXX Sendmail compatibility: don't expose our first body line.
      */
@@ -2079,6 +2106,8 @@ static void milter8_eob(void *ptr)
     MILTER_MSG_CONTEXT *msg_ctx = (MILTER_MSG_CONTEXT *) ptr;
     MILTER8 *milter = msg_ctx->milter;
 
+    if (MILTER8_MESSAGE_DONE(milter, msg_ctx))
+	return;
     if (msg_verbose)
 	msg_info("%s: eob milter %s", myname, milter->m.name);
     msg_ctx->resp =
@@ -2144,7 +2173,7 @@ static const char *milter8_message(MILTER *m, VSTREAM *qfile,
 
 	/*
 	 * XXX When the message (not MIME body part) does not end in CRLF
-	 * (i.e. the last record was REC_TYPE_CONT), do we send CRLF
+	 * (i.e. the last record was REC_TYPE_CONT), do we send a CRLF
 	 * terminator before triggering the end-of-body condition?
 	 */
 	for (;;) {
@@ -2164,9 +2193,7 @@ static const char *milter8_message(MILTER *m, VSTREAM *qfile,
 		msg_ctx.resp = "450 4.3.0 Queue file write error";
 		break;
 	    }
-	    if (msg_ctx.resp != 0)
-		break;
-	    if (milter->state != MILTER8_STAT_MESSAGE)
+	    if (MILTER8_MESSAGE_DONE(milter, &msg_ctx))
 		break;
 	    if (rec_type != REC_TYPE_NORM && rec_type != REC_TYPE_CONT)
 		break;
